import { config } from "dotenv";
config();

import { col } from "./src/db";
import { Transaction } from "./src/lib/types";
import { generateMetadata, loadRules } from "./src/lib/rules";
import {
  FintAccount,
  FintRawTransaction,
  RawTransactionPayPal,
} from "./src/db/types.fint";
import { getAmount, getDate } from "./src/lib/data";

import { writeFile, readFile } from "node:fs/promises";

import { mkdirp } from "mkdirp";

const accounts = await col<FintAccount>("accounts").find().toArray();
const rules = await loadRules();

const targetDir: Record<string, string> = JSON.parse(
  await readFile("mapping.json").then((f) => f.toString())
);

for (const account of accounts) {
  if (targetDir[account._id]) {
    const rawTransactions = await col<FintRawTransaction>("raw_transactions")
      .find({
        "_id.accountId": account._id,
        // PayPal
        "data.Balance Impact": { $ne: "Memo" }, // Invoices
        "data.Status": { $ne: "Refused" }, // Failed transactions
      })
      .toArray();

    const mergedRaw: Record<string, FintRawTransaction[]> = {};
    rawTransactions.forEach(
      (transaction) =>
        (mergedRaw[transaction.hash] = [
          ...(mergedRaw[transaction.hash] ?? []),
          transaction as FintRawTransaction,
        ])
    );

    let transactionsList: Transaction[] = Object.keys(mergedRaw)
      .map((hash) => {
        const rawEntries = mergedRaw[hash];
        const transaction: Transaction = {
          _id: String(),
          accountId: account._id,

          hash,
          date: getDate(rawEntries),
          amount: getAmount(rawEntries),
          metadata: generateMetadata(rawEntries, rules),
          pending: !rawEntries.find((x) => !x.pending),
        };

        (transaction as any).raw = rawEntries;
        return transaction;
      })
      .toSorted((a, b) => +a.date - +b.date);

    console.info(
      targetDir[account._id],
      "balance: £",
      transactionsList.reduce(
        (acc, transaction) => acc + parseFloat(transaction.amount) * 100,
        0
      ) / 100
    );

    const counter: Record<string, number> = {};
    function count(yyyymmdd: string) {
      counter[yyyymmdd] = (counter[yyyymmdd] ?? 0) + 1;
      return counter[yyyymmdd];
    }

    for (const transaction of transactionsList) {
      if (
        ((transaction as any).raw as FintRawTransaction[]).find(
          (transaction) =>
            transaction._id.source === "manual_paypal" &&
            [
              // Temporary holds on balance
              "Account Hold for Open Authorisation",
              "Reversal of General Account Hold",
              "Payment Hold",
              "Payment Release",
              "Payment Review Hold",
              "Payment Review Release",
            ].includes((transaction.data as RawTransactionPayPal).Type)
        )
      ) {
        continue;
      }

      const yyyymm =
        transaction.date.getUTCFullYear().toString() +
        "-" +
        (transaction.date.getUTCMonth() + 1).toString().padStart(2, "0");

      const yyyymmdd =
        transaction.date.getUTCFullYear().toString() +
        (transaction.date.getUTCMonth() + 1).toString().padStart(2, "0") +
        transaction.date.getUTCDate().toString().padStart(2, "0");

      const idx = count(yyyymmdd.toString());
      const dir = `${targetDir[account._id]}/${
        transaction.pending ? "pending" : yyyymm
      }`;

      await mkdirp(dir);
      await writeFile(
        `${dir}/${yyyymmdd}-${idx}.uff1.json`,
        JSON.stringify(
          {
            time: transaction.date.toISOString(),
            net: parseFloat(transaction.amount).toFixed(2),
            data: transaction,
          } as UFF1,
          undefined,
          2
        ) +
          ((
            await readFile(`${dir}/${yyyymmdd}-${idx}.uff1.json`)
              .then((f) => f.toString())
              .catch(() => "\n")
          ).endsWith("\n")
            ? "\n"
            : "")
      );
    }

    for (let i = -60 + 1; i <= 0; i++) {
      const boundLeft = new Date();
      boundLeft.setDate(1);
      boundLeft.setHours(0);
      boundLeft.setMinutes(0);
      boundLeft.setSeconds(0);
      boundLeft.setMonth(boundLeft.getMonth() + i);

      const boundRight = new Date(boundLeft);
      boundRight.setMonth(boundRight.getMonth() + 1);

      // console.info(
      //   `${boundLeft.getFullYear()}-${boundLeft.getMonth() + 1}`,
      //   transactionsList
      //     .filter(
      //       (entry) => entry.date >= boundLeft && entry.date < boundRight
      //     )
      //     .toSorted((a, b) => +a.date - +b.date)
      //     .map((entry) => entry.amount)
      // );

      // if (boundLeft.getFullYear() === 2024 && boundRight.getMonth() === 5)
      //   console.info(
      //     inspect(
      //       transactionsList
      //         .filter(
      //           (entry) =>
      //             entry.date >= boundLeft && entry.date < boundRight
      //         )
      //         .toSorted((a, b) => +a.date - +b.date),
      //       { colors: true }
      //     )
      //   );
    }
  }
}

process.exit(0);
