import { FintRawTransaction } from "../db/types.fint";
import { TransactionMetadata } from "./types";
import { readFile } from "node:fs/promises";

type Matcher =
  | {
      type: "substring";
      keys: string[];
      substr: string;
    }
  | {
      type: "eq";
      keys: string[];
      str: string;
    }
  | {
      type: "and";
      matchers: Matcher[];
    }
  | {
      type: "or";
      matchers: Matcher[];
    }
  | {
      type: "not";
      matcher: Matcher;
    };

type Rule = {
  match: Matcher;
  set: Partial<TransactionMetadata>;
};

export function loadRules() {
  try {
    return readFile("rules.json").then(
      (f) => JSON.parse(f.toString()) as Rule[]
    );
  } catch (err) {
    return [];
  }
}

function jsonKey(obj: any, key: string) {
  if (typeof obj === "undefined") return undefined;
  if (key.length === 0) return obj;
  if (typeof obj === "object") {
    const segments = key.split(".");
    return jsonKey(obj[segments.shift()!], segments.join("."));
  } else {
    return undefined;
  }
}

function jsonKeyI(obj: any, key: string) {
  const arrValues = jsonKey(obj, key + "Array");
  return [
    jsonKey(obj, key),
    ...(Array.isArray(arrValues) ? arrValues : []),
  ].filter((x) => x);
}

function isMatch(obj: any, matcher: Matcher): boolean {
  switch (matcher.type) {
    case "substring": {
      const { keys, substr } = matcher;

      return keys.some((key) => {
        const values = jsonKeyI(obj, key);

        for (const value of values) {
          if (typeof value === "string") {
            if (value.toLowerCase().includes(substr.toLowerCase())) {
              return true;
            }
          }
        }
      });
    }
    case "eq": {
      const { keys, str } = matcher;

      return keys.some((key) => {
        const values = jsonKeyI(obj, key);

        for (const value of values) {
          if (typeof value === "string") {
            if (value.toLowerCase() === str.toLowerCase()) {
              return true;
            }
          }
        }
      });
    }
    case "and":
      return matcher.matchers.every((matcher) => isMatch(obj, matcher));
    case "or":
      return matcher.matchers.some((matcher) => isMatch(obj, matcher));
    case "not":
      return !isMatch(obj, matcher.matcher);
  }
}

export function generateMetadata(
  rawEntries: FintRawTransaction[],
  rules: Rule[]
): TransactionMetadata {
  let metadata: TransactionMetadata = {
    category: "uncategorised",
  };

  for (const entry of rawEntries) {
    for (const rule of rules) {
      if (isMatch(entry.data, rule.match)) {
        metadata = {
          ...metadata,
          ...rule.set,
        };
      }
    }
  }

  return metadata;
}
