import { Account, FinancialRecord, UFF, UFF1 } from "./lib/uff";
import { resolve } from "path";
import { writeFile, unlink } from "fs/promises";
import stringify from "json-stringify-deterministic";

const record = await FinancialRecord.fromDirectory("../financial-record");

class MetaHelper {
  #record: FinancialRecord;

  constructor(record: FinancialRecord) {
    this.#record = record;
  }

  transactionsByProperties(accountId: string, raw: Partial<UFF1>) {
    const account = this.#record.getAccount(accountId);
    return account.allTransactions.filter((transaction) =>
      Object.keys(raw).every((key) => transaction[key] === raw[key])
    );
  }

  transactionByProperties(accountId: string, raw: Partial<UFF1>) {
    const candidates = this.transactionsByProperties(accountId, raw);

    if (candidates.length > 1)
      throw new Error(
        "Found more than 1 candidate for match! " + JSON.stringify(raw)
      );

    return candidates[0];
  }

  transactionsByRawString(accountId: string, raw: string) {
    const account = this.#record.getAccount(accountId);
    return account.allTransactions.filter((transaction) =>
      transaction.rawDataAsString.includes(raw)
    );
  }
}

class FinancialRecordAccountTransactionIterator {
  readonly meta: MetaHelper;
  #transactions: UFF[];
  #index: number;

  constructor(record: FinancialRecord, account: Account) {
    this.meta = new MetaHelper(record);
    this.#transactions = account.allTransactions;
    this.#index = 0;
  }

  get item() {
    return this.#transactions[this.#index];
  }

  next() {
    this.#index++;
    return this.item;
  }

  reset() {
    this.#index = 0;
    return this.item;
  }
}

let rootModule;
try {
  rootModule = await import(resolve(record.filePath, "rules.js"));
} catch (err) {
  console.error("No root rules specified!");
}

for (const id of record.accountIds) {
  // Categorisation routine
  const account = record.getAccount(id);
  const iterator = new FinancialRecordAccountTransactionIterator(
    record,
    account
  );

  let accountModule;
  try {
    accountModule = await import(resolve(account.filePath, "rules.js"));
  } catch (err) {
    console.error(
      "No account rules for",
      id + ", I will only apply root rules!"
    );
  }

  let item = iterator.item;
  while (item) {
    // Mark dummy transactions
    if (item.netFixed === 0) {
      item.type ??= "Dummy";
      item.category ??= "Misc";
    }

    // Manually crafted rules
    rootModule?.metadata(item, iterator.meta);
    accountModule?.metadata(item, iterator.meta);

    item = iterator.next();
  }
}

console.log();

for (const id of record.accountIds) {
  const account = record.getAccount(id);
  const iterator = new FinancialRecordAccountTransactionIterator(
    record,
    account
  );

  let item = iterator.item;
  while (item) {
    // Module: Apply PayPal currency conversion chain linking (fint-JSON)
    // May need adapting to other data formats, but we assume that:
    // - PayPal activity export was used
    // - The data is in JSON format
    // - PayPal account is prefixed paypal_
    if (id.startsWith("paypal_")) {
      const data = item.rawDataAsString;

      // Determine if currency conversion
      if (data.includes("General Currency Conversion")) {
        // Extract the reference ID
        const match = /"Reference Txn ID": *"([A-Z0-9]+?)"/.exec(data);
        if (match) {
          const [, txn] = match;

          // Scan all transactions for matching transactions
          // (we are looking for opposite side AND the actual transaction)
          for (const account of record.accountIds.filter(
            (other) => id !== other && other.startsWith("paypal_")
          )) {
            const matches = iterator.meta.transactionsByRawString(account, txn);

            // We are from perspective of source, apply from here:
            if (matches.length === 2) {
              item.type = "Conversion";
              item.category = "Currency Conversion";
              item.payee = "PayPal";

              item.links = matches.map((transaction) => ({
                type: "Direct",
                accountId: account,
                transactionId: transaction.id,
              }));

              for (const transaction of matches) {
                const other = matches.find(
                  (other) => other.id !== transaction.id
                )!;

                transaction.links = [
                  {
                    type: "Direct",
                    accountId: account,
                    transactionId: other.id,
                  },
                  {
                    type: "Direct",
                    accountId: id,
                    transactionId: item.id,
                  },
                ];

                if (
                  transaction.rawDataAsString.includes(
                    "General Currency Conversion"
                  )
                ) {
                  transaction.type = "Conversion";
                  transaction.category = "Currency Conversion";
                  transaction.payee = "PayPal";
                }
              }

              break;
            } else if (matches.length !== 1) {
              // throw new Error("Missing data!");
            }
          }
        }
      }
    }

    // Module: Apply Revolut currency conversion chain linking (fint-JSON)
    // May need adapting to other data formats, but we assume that:
    // - GoCardless export was used
    // - Revolut accounts are revolut_${currency}
    if (id.startsWith("revolut_")) {
      /// && !item.links) {
      const data = item.rawDataAsString;

      // Determine if currency conversion
      if (data.includes("EXCHANGE")) {
        // Extract the transaction ID
        const txidMatch =
          /"transactionId": *"([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})"/.exec(
            data
          );

        if (txidMatch) {
          const [, txid] = txidMatch;

          // Blind search for the correct transaction
          for (const account of record.accountIds.filter(
            (other) => id !== other && other.startsWith("revolut_")
          )) {
            const matches = iterator.meta.transactionsByRawString(
              account,
              txid
            );

            if (matches.length > 1)
              throw new Error("Somehow matched more than once!");

            const match = matches[0];
            if (match) {
              item.type = "Conversion";
              item.category = "Currency Conversion";
              item.payee = "Revolut";

              match.type = "Conversion";
              match.category = "Currency Conversion";
              match.payee = "Revolut";

              item.links = [
                {
                  type: "Direct",
                  accountId: account,
                  transactionId: match.id,
                },
              ];

              match.links = [
                {
                  type: "Direct",
                  accountId: id,
                  transactionId: item.id,
                },
              ];
            }
          }
        }
      }
    }

    item = iterator.next();
  }

  // Print stats
  const all = account.allTransactions;
  const missingTransactions = all.filter(
    (x) => !x.category && !x.type && !x.filePath.includes("pending")
  );

  console.info(
    `${id} has balance: £`,
    account.balance,
    "and has",
    missingTransactions.length,
    "transactions missing category and type out of",
    all.length,
    "transactions."
  );

  if (missingTransactions.length) {
    await writeFile(
      `${id}_missing.json`,
      stringify(
        missingTransactions.map((x) => ({
          id: x.filePath
            .split("/")
            .pop()
            ?.replace(/.uff1.json$/, ""),
          net: x.net,
          raw: x.rawData,
        })),
        { space: "  " }
      ) + "\n"
    );
  } else {
    try {
      await unlink(`${id}_missing.json`);
    } catch (_) {}
  }
}

await record.commit();

// Use hints from data to do balance validation
// for (const id of record.accountIds) {
//   const account = record.getAccount(id);
//   const iterator = new FinancialRecordAccountTransactionIterator(
//     record,
//     account
//   );
//
//   let item = iterator.item;
//   while (item) {
//     // Module: Check balance matches the cumulative sum in our database
//     // May need adapting to other data formats, but we assume that:
//     // - PayPal activity export was used
//     // - The data is in JSON format
//     // - PayPal account is prefixed paypal_
//     let sum = 0,
//       lastBalance = 0;
//     if (id.startsWith("paypal_")) {
//       const data = item.rawDataAsString;
//       const balMatch = /"Balance": *"(-?(\d+,)*\d+\.\d+)"/.exec(data);
//       if (!balMatch)
//         throw "No 'Balance' field in " + item.id + " while processing " + id;
//
//       lastBalance = parseInt(balMatch[1].replace(/,|\./g, ""));
//       sum += item.netFixed;
//     }
//
//     item = iterator.next();
//   }
// }

// Check strict rules about data
for (const id of record.accountIds) {
  const account = record.getAccount(id);

  for (const transaction of account.allTransactions) {
    switch (transaction.type) {
      case "Compensation":
      case "Expense":
        if (transaction.netFixed > 0) {
          console.error(
            transaction,
            transaction.net,
            transaction.type,
            transaction.filePath,
            transaction.filePath2,
            transaction.rawData
          );
          throw "Invalid transaction type! (expense)";
        }
        break;
      case "Income":
      case "Interest":
      case "Reimbursement":
        if (transaction.netFixed < 0) {
          console.error(
            transaction,
            transaction.net,
            transaction.type,
            transaction.filePath,
            transaction.filePath2,
            transaction.rawData
          );
          throw "Invalid transaction type! (income)";
        }
        break;
      case "Transfer":
        if (
          !transaction.links &&
          (!transaction.category ||
            ![
              // no corresponding transaction
              "ATM",
              "Crypto",
              "Deposit",
              // may not be up to date
              "Savings",
            ].includes(transaction.category!)) &&
          transaction.payee === "Self"
        ) {
          console.error(
            transaction,
            transaction.net,
            transaction.type,
            transaction.filePath,
            transaction.filePath2,
            transaction.rawData
          );
          throw "Invalid transaction type! (transfer)";
        }
        break;
    }
  }
}
