import stringify from "json-stringify-deterministic";

import { readFile, readdir, stat, writeFile } from "node:fs/promises";
import { resolve } from "node:path";

/**
 * Type 1 UFF file
 * This is the basic information about a transaction
 * and contains the raw imported data about it.
 */
export type UFF1 = {
  /**
   * ISO8601 Timestamp
   */
  time: string;

  /**
   * Net balance change
   */
  net: string;

  /**
   * Raw information about this transaction
   *
   * Can be in any data format
   */
  data: any;
};

/**
 * Type 2 UFF file
 * This is a transaction descriptor file, it contains
 * metadata generated from the information in UFF1.
 */
export type UFF2 = {
  /**
   * Description
   */
  description?: string;

  /**
   * Payee
   */
  payee?: string;

  /**
   * Transaction type
   *
   * Dummy:         Zero net transaction
   * Transfer:      Money going between accounts
   * Conversion:    Special-case of transfer where currency is converted
   *
   * [Income]
   * Income:        Money coming in considered income
   * Interest:      Interest on balance
   * Reimbursement: Money coming in considered a reimbursement
   *
   * [Income / Expense]
   * Loan:          Money coming in considered a loan
   *
   * [Expense]
   * Compensation : Money going out considered compensation
   * Expense:       Money going out
   */
  type?:
    | "Transfer"
    | "Conversion"
    | "Income"
    | "Interest"
    | "Loan"
    | "Reimbursement"
    | "Compensation"
    | "Expense"
    | "Dummy";

  /**
   * Category of the transaction
   */
  category?: string;

  /**
   * Links to other transactions
   */
  links?: {
    /**
     * Type of link to other transaction
     *
     * Direct:   Corresponding cause/effect transaction that this links to, e.g. a transfer
     * Indirect: Corresponding transactions that this transaction is connected to
     */
    type: "Direct" | "Indirect";

    /**
     * Account ID in record
     */
    accountId: string;

    /**
     * Transaction ID in account
     */
    transactionId: string;

    /**
     * An optional comment on why this is linked
     */
    comment?: string;
  }[];

  /**
   * Pool this money belongs to
   */
  pool?: string;

  /**
   * Invoice ID
   */
  invoice?: string;

  /**
   * Marker to prevent edits
   */
  __immutable?: boolean;
};

/**
 * Loaded UFF data
 */
export class UFF {
  readonly id: string;
  readonly filePath: string;
  #transaction: UFF1;
  #metadata: UFF2;
  #pendingWrite: boolean;

  constructor(
    filePath: string,
    transaction: UFF1,
    metadata: UFF2,
    createdMetaFile: boolean
  ) {
    this.id = filePath
      .split("/")
      .pop()!
      .replace(/.uff1.json$/, "");
    this.filePath = filePath;
    this.#transaction = transaction;
    this.#metadata = metadata;
    this.#pendingWrite = createdMetaFile;
  }

  get timeStr() {
    return this.#transaction.time;
  }

  get time() {
    return new Date(this.#transaction.time);
  }

  get net() {
    return this.#transaction.net;
  }

  get isIncome() {
    const isIncome =
      this.type === "Income" ||
      this.type === "Interest" ||
      this.type === "Reimbursement" ||
      (this.type === "Loan" && this.netFixed > 0);

    if (isIncome && this.netFixed < 0) throw new Error("Data mis-match!");
    return isIncome;
  }

  get isExpense() {
    const isExpense =
      this.type === "Expense" ||
      this.type === "Compensation" ||
      (this.type === "Loan" && this.netFixed < 0);

    if (isExpense && this.netFixed > 0) throw new Error("Data mis-match!");
    return isExpense;
  }

  static RE_NORMALISED_VALUE = /^-?(0|[1-9][0-9]*)\.[0-9]{2}$/;

  get netFixed() {
    if (!this.net.match(UFF.RE_NORMALISED_VALUE)) {
      throw new Error("Value is not normalised: " + this.net);
    }

    return parseInt(this.net.replace(".", ""));
  }

  get rawData() {
    return this.#transaction.data;
  }

  get rawDataAsString() {
    return typeof this.rawData !== "string"
      ? stringify(this.rawData)
      : this.rawData;
  }

  get description() {
    return this.#metadata?.description;
  }

  set description(value: string | undefined) {
    if (value) {
      this.#metadata.description = value;
    } else {
      delete this.#metadata.description;
    }

    this.#pendingWrite = true;
  }

  get payee() {
    return this.#metadata?.payee;
  }

  set payee(value: string | undefined) {
    if (value) {
      this.#metadata.payee = value;
    } else {
      delete this.#metadata.payee;
    }

    this.#pendingWrite = true;
  }

  get type() {
    return this.#metadata?.type;
  }

  set type(value: UFF2["type"]) {
    if (value) {
      this.#metadata.type = value;
    } else {
      delete this.#metadata.type;
    }

    this.#pendingWrite = true;
  }

  get category() {
    return this.#metadata?.category;
  }

  set category(value: string | undefined) {
    if (value) {
      this.#metadata.category = value;
    } else {
      delete this.#metadata.category;
    }

    this.#pendingWrite = true;
  }

  get links() {
    return this.#metadata?.links;
  }

  set links(value: UFF2["links"]) {
    if (value) {
      this.#metadata.links = value;
    } else {
      delete this.#metadata.links;
    }

    this.#pendingWrite = true;
  }

  get pool() {
    return this.#metadata?.pool;
  }

  set pool(pool: string | undefined) {
    if (pool) {
      this.#metadata.pool = pool;
    } else {
      delete this.#metadata.pool;
    }

    this.#pendingWrite = true;
  }

  get invoice() {
    return this.#metadata.invoice;
  }

  get immutable() {
    return this.#metadata.__immutable ?? false;
  }

  get filePath2() {
    return this.filePath.replace(/.uff1.json$/, ".uff2.json");
  }

  static async fromFile(filePath: string, uff2?: boolean) {
    const transaction = JSON.parse(
      await readFile(filePath).then((f) => f.toString())
    );

    let metadata = {};
    const filePath2 = filePath.replace(/.uff1.json$/, ".uff2.json");
    if (typeof uff2 !== "boolean" || uff2) {
      try {
        metadata = JSON.parse(
          await readFile(filePath2).then((f) => f.toString())
        );
      } catch (err) {
        if (uff2) {
          throw err;
        } else {
          uff2 = false;
        }
      }
    }

    return new UFF(filePath, transaction, metadata, uff2 === false);
  }

  async commit() {
    if (this.#pendingWrite) {
      await this.#writeMetadata();
    }
  }

  async #writeMetadata() {
    await writeFile(
      this.filePath2,
      stringify(this.#metadata, { space: "  " }) + "\n"
    );
  }
}

const RE_MONTH = /^\d{4}-\d{2}$|^pending$/;

/**
 * Loaded UFF account data
 */
export class Account {
  readonly filePath: string;
  #transactions: UFF[];

  constructor(filePath: string, transactions: UFF[]) {
    this.filePath = filePath;
    this.#transactions = transactions;
  }

  static async fromDirectory(directory: string) {
    const list: UFF[] = [];
    for (const month of await readdir(directory).then((arr) =>
      arr.toSorted()
    )) {
      if (!RE_MONTH.test(month)) continue;

      const file = await stat(resolve(directory, month));
      if (!file.isDirectory()) continue;

      const transactions = await readdir(resolve(directory, month));
      for (const transaction of transactions.toSorted()) {
        if (!transaction.endsWith(".uff1.json")) continue;

        const uff = await UFF.fromFile(resolve(directory, month, transaction));
        list.push(uff);
      }
    }

    return new Account(directory, list);
  }

  get allTransactions() {
    return [...this.#transactions];
  }

  get balance() {
    return this.balanceFixed / 100;
  }

  get balanceFixed() {
    return this.#transactions.reduce((acc, uff) => acc + uff.netFixed, 0);
  }

  async commit() {
    for (const transaction of this.#transactions) {
      transaction.commit();
    }
  }
}

/**
 * Loaded UFF repository
 */
export class FinancialRecord {
  readonly filePath: string;
  #accounts: Record<string, Account>;

  constructor(filePath: string, accounts: Record<string, Account>) {
    this.filePath = filePath;
    this.#accounts = accounts;
  }

  static async fromDirectory(directory: string) {
    const accounts: Record<string, Account> = {};
    for (const accountId of await readdir(directory)) {
      if (accountId === ".git") continue;

      const file = await stat(resolve(directory, accountId));
      if (!file.isDirectory()) continue;

      accounts[accountId] = await Account.fromDirectory(
        resolve(directory, accountId)
      );
    }

    return new FinancialRecord(directory, accounts);
  }

  getAccount(id: string) {
    return this.#accounts[id];
  }

  get accountIds() {
    return Object.keys(this.#accounts);
  }

  async commit() {
    for (const accountId of this.accountIds) {
      await this.getAccount(accountId).commit();
    }
  }

  get allTransactions() {
    return this.accountIds.flatMap((id) => this.getAccount(id).allTransactions);
  }
}
