import { UFF } from "./uff";

export class TransactionUtil {
  readonly transactions: UFF[];

  constructor(transactions: UFF[]) {
    this.transactions = transactions;
  }

  satisfy(f: undefined | ((t: UFF) => boolean)) {
    if (!f) return this;
    return new TransactionUtil(this.transactions.filter(f));
  }

  payee(payee: string | string[]) {
    return new TransactionUtil(
      Array.isArray(payee)
        ? this.transactions.filter((t) => payee.includes(t.payee!))
        : this.transactions.filter((t) => t.payee === payee)
    );
  }

  category(category: string | string[]) {
    return new TransactionUtil(
      Array.isArray(category)
        ? this.transactions.filter((t) => category.includes(t.category!))
        : this.transactions.filter((t) => t.category === category)
    );
  }

  pool(pool?: string) {
    return new TransactionUtil(
      pool
        ? this.transactions.filter((t) => t.pool === pool)
        : this.transactions.filter((t) => !t.pool)
    );
  }

  ukTaxYear(startingYear: number) {
    const startDate = new Date(`${startingYear}-04-06T00:00:00Z`);
    const endDate = new Date(`${startingYear + 1}-04-06T00:00:00Z`);

    return new UkTaxTransactionUtil(
      this.transactions.filter((t) => t.time >= startDate && t.time < endDate)
    );
  }

  year(year: number, q?: number) {
    return new TransactionUtil(
      this.transactions.filter(
        (t) =>
          t.time.getFullYear() === year &&
          (q ? Math.ceil((t.time.getMonth() + 1) / 3) === q : true)
      )
    );
  }

  untilYear(year: number, q?: number) {
    return new TransactionUtil(
      this.transactions.filter(
        (t) =>
          t.time.getFullYear() < year ||
          (t.time.getFullYear() === year &&
            (q ? Math.ceil((t.time.getMonth() + 1) / 3) <= q : true))
      )
    );
  }

  get netFixed() {
    return this.transactions.reduce((acc, t) => acc + t.netFixed, 0);
  }

  get net() {
    return this.netFixed / 100;
  }

  get netAsString() {
    return this.net.toFixed(2);
  }

  get income() {
    return new TransactionUtil(this.transactions.filter((t) => t.isIncome));
  }

  get spend() {
    return new TransactionUtil(this.transactions.filter((t) => t.isExpense));
  }

  get misc() {
    return new TransactionUtil(
      this.transactions.filter((t) => !t.isIncome && !t.isExpense)
    );
  }

  get first() {
    return this.transactions.toSorted((a, b) => +a.time - +b.time)[0];
  }

  get last() {
    return this.transactions.toSorted((a, b) => +a.time - +b.time)[
      this.transactions.length - 1
    ];
  }
}

export class UkTaxTransactionUtil extends TransactionUtil {
  get taxableIncome() {
    return this.satisfy(
      (t) =>
        (t.type === "Income" &&
          t.category !== "Gift" &&
          t.category !== "Tax-free") ||
        t.type === "Interest"
    );
  }

  get incomeTaxFixed() {
    const net = this.taxableIncome.netFixed;

    // TODO: make a table that considers the tax year, etc
    const basicBracket = Math.min(
      Math.max(0, net - 12_571_00),
      50_271_00 - 12_571_00
    );

    const higherBracket = Math.min(
      Math.max(0, net - 50_271_00),
      125_140_00 - 12_571_00
    );

    const additionalBracket = Math.min(Math.max(0, net - 125_140_00));

    return basicBracket * 0.2 + higherBracket * 0.4 + additionalBracket * 0.45;
  }

  get incomeTax() {
    return this.incomeTaxFixed / 100;
  }

  // TODO: https://www.gov.uk/self-employed-national-insurance-rates
}
